/*
    CalcWindow.cpp

    Copyright  1996 Levin-Delson Software.  All Rights Reserved.

    $Header: /Bedazzle/dvlp/Calculator/RCS/CalcWindow.cpp,v 1.8 1996/09/16 21:07:50 adamld Exp $
*/

#include "CalcWindow.h"

CalcWindow :: CalcWindow( BRect frame, char *title, BHandler *bHandler ) : BWindow( frame, title, B_TITLED_WINDOW, ( B_NOT_RESIZABLE | B_NOT_ZOOMABLE | B_WILL_ACCEPT_FIRST_CLICK | B_WILL_FLOAT ) )
{
    calcMessenger = new BMessenger( bHandler );

	mdFilter = new MouseDownFilter( B_ANY_DELIVERY, B_ANY_SOURCE, B_MOUSE_DOWN );

	kdFilter = new KeyDownFilter( B_ANY_DELIVERY, B_ANY_SOURCE, B_KEY_DOWN );

	if ( TRUE == Lock() )
	{
		if ( NULL != mdFilter )
		{
				AddCommonFilter( mdFilter );
		}
		
		if ( NULL != kdFilter )
		{
				AddCommonFilter( kdFilter );
		}
		
		Unlock();
	}

    numLocked = 0UL;
}


CalcWindow :: ~CalcWindow()
{
	if ( TRUE == Lock() )
	{
		if ( NULL != kdFilter )
		{
			RemoveCommonFilter( kdFilter );
			delete( kdFilter );
			kdFilter = NULL;
		}

		if ( NULL != mdFilter )
		{
			RemoveCommonFilter( mdFilter );
			delete( mdFilter );
			mdFilter = NULL;
		}
		
		Unlock();
	}
	
    if ( NULL != calcMessenger )
    {
        delete( calcMessenger );
		calcMessenger = NULL;
    }
}


/* The window receives several types of messages:

    Update messages from the Calculator app which contain a new display string.

    Requests to Enable/Disable buttons.

    Clipboard requests.
*/
void CalcWindow :: MessageReceived( BMessage *bMessage )
{
	char    *objectName;
    Display *display;
    BButton *bButton;
    char    *text;
    bool     enabled;


    if ( NULL != bMessage )
    {
        switch ( bMessage->what )
        {
            case ( CALC_MESSAGE_SETTEXT ):
                /*
                    The display needs to be updated.  Copy the string in the message to
                    the display BStringView.
                */
                objectName = (char *)bMessage->FindString( CALC_MESSAGEID_OBJECTNAME );
                if ( ( NULL != objectName ) && ( B_NO_ERROR <= bMessage->Error() ) )
                {
	                text = (char *)bMessage->FindString( CALC_MESSAGEID_TEXT );
	                if ( ( NULL != text ) && ( B_NO_ERROR <= bMessage->Error() ) )
	                {
	                    display = cast_as( FindView( objectName ), Display );
	                    if ( NULL != display )
	                    {
	                        display->SetText( text );
	                    }
					}
                }
            break;

            case ( CALC_MESSAGE_SETENABLED ):
                // Enable/Disable the named button according to the enable variable.
                objectName = (char *)bMessage->FindString( CALC_MESSAGEID_OBJECTNAME );
                if ( ( NULL != objectName ) && ( B_NO_ERROR <= bMessage->Error() ) )
                {
                    enabled = bMessage->FindBool( CALC_MESSAGEID_ENABLED );
                    if ( B_NO_ERROR <= bMessage->Error() )
                    {
                        bButton = cast_as( FindView( objectName ), BButton );
                        if ( NULL != bButton )
                        {
                            bButton->SetEnabled( enabled );
                        }
                    }
                }
            break;

            case ( B_CUT ):
            case ( B_COPY ):
            case ( B_PASTE ):
                // Tell the calculator app to perform the clipboard function.
                (void) calcMessenger->SendMessage( bMessage->what );
            break;

            default:
                // If we don't recognize the message, pass it to the parent class.
                BWindow::MessageReceived( bMessage );
            break;
        }
    }
}


/*
    WindowActivated is used to turn on the Num Lock on activation and
    reset it on deactivation.
*/
void CalcWindow :: WindowActivated( bool active )
{
    ulong modifierBits;

    // Grab the current state of the "locks".
    modifierBits = modifiers();

    if ( TRUE == active )
    {
        // Window just became active.  Save state of Num Lock and them lock it.
        numLocked = modifierBits & B_NUM_LOCK;

        set_keyboard_locks( B_NUM_LOCK | ( modifierBits & ( B_CAPS_LOCK | B_SCROLL_LOCK ) ) );
    }
    else
    {
        // Window just became inactive.  Restore saved state of Num Lock.
        set_keyboard_locks( numLocked | ( modifierBits & ( B_CAPS_LOCK | B_SCROLL_LOCK ) ) );
    }

    // Pass this on just in case any other activity is called for.
    BWindow::WindowActivated( active );
}


// The window forwards a message to the app.
long CalcWindow :: forward( ulong what )
{
	return( calcMessenger->SendMessage( what ) );
}


bool CalcWindow :: QuitRequested( void )
{
    // Deactivate the window.
    WindowActivated( FALSE );

    // Tell the app that the user asked to quit.
    (void) calcMessenger->SendMessage( B_QUIT_REQUESTED );

    return ( TRUE );
}


/*
	MouseDownFilter manages a PopUp menu that is presented when the secondary (menu) mousebutton
	is pressed.  The filter will ask the window to forward a message containing the menu action
	to the app.
*/
MouseDownFilter :: MouseDownFilter( message_delivery delivery, message_source source, ulong command ) : BMessageFilter( delivery, source, command )
{
    BMenuItem      *item;


    menu = new BPopUpMenu( "Clipboard", FALSE, FALSE );
    if ( NULL != menu )
    {
        item = new BMenuItem( "Cut", new BMessage( B_CUT ), 'X' );
        if ( NULL != item )
        {
            menu->AddItem( item );
        }

        item = new BMenuItem( "Copy", new BMessage( B_COPY ), 'C' );
        if ( NULL != item )
        {
            menu->AddItem( item );
        }

        item = new BMenuItem( "Paste", new BMessage( B_PASTE ), 'V' );
        if ( NULL != item )
        {
            menu->AddItem( item );
        }
    }
}


MouseDownFilter :: ~MouseDownFilter()
{
	if ( NULL != menu )
	{
		delete( menu );
		menu = NULL;
	}
}


/*
	MouseDownFilter:
	
		Activates the window on any mouse activity.
		
		If the secondary (menu) mousebutton is pressed, the function presents the menu to
		the user and forwards the requested menu action to the app.

*/
filter_result MouseDownFilter :: Filter( BMessage *message, BHandler **target )
{
	filter_result  result = B_DISPATCH_MESSAGE;
	BView		  *view = NULL;
	CalcWindow    *calcWindow;
	ulong		   buttons;
	BPoint		   cursor;
	BView		  *parentView;
    BMenuItem     *menuItem;
	
	
	if ( NULL != target )
	{
		view = cast_as( *target, BView );
	}
	
	if ( ( NULL != message ) && ( NULL != view ) )
	{
		calcWindow = cast_as( view->Window(), CalcWindow );
		if ( NULL != calcWindow )
		{
			// Activate the window if it isn't already active.
			if ( FALSE == calcWindow->IsActive() )
			{
			    calcWindow->Activate( TRUE );
			}
		
			// Find out which buttons went down.
			buttons = (ulong)message->FindLong( "buttons" );
			
			if ( B_NO_ERROR <= message->Error() )
			{
				// If the menu button is down, bring up the menu.
				if ( B_SECONDARY_MOUSE_BUTTON == ( buttons & B_SECONDARY_MOUSE_BUTTON ) )
				{
					/*
						We want to bring the menu up under the cursor.  The message passed to us
						has a cursor location relative to the View's frame.  We need a position
						relative to the Window's frame, so we'll walk up the chain of Views to
						the top view (whose frame covers the window's frame) then ask it for the
						cursor's location.
					*/ 
					while ( NULL != ( parentView = view->Parent() ) )
					{
						view = parentView;
					}
					
					// Get the current mouse info.
					view->GetMouse( &cursor, &buttons, FALSE );
					
					calcWindow->ConvertToScreen( &cursor );
					
					// Display the menu then obtain the menuitem selected.
					menuItem = menu->Go( cursor );
					if ( NULL != menuItem )
					{
						// Have the window forward the action to the app.
						(void) calcWindow->forward( menuItem->Command() );
					}
				
					// Prevent the mousedown menu event from selecting a button.
					result = B_SKIP_MESSAGE;
				}
			}
		}
	}
	
	return( result );
}
			
			
/*
	KeyDownFilter
	
	Filters KeyDown messages and forwards keypresses to the application.
	(Keypresses, unlike button events, can't be targetted directly to the app,
	 so they are forwarded there when received by the window.  Note that some
	 additional keys [e.g. upper and lower case] are supported for some buttons.)
*/

KeyDownFilter :: KeyDownFilter( message_delivery delivery, message_source source, ulong command ) : BMessageFilter( delivery, source, command )
{
}


filter_result KeyDownFilter :: Filter( BMessage *message, BHandler **target )
{
	filter_result  result = B_DISPATCH_MESSAGE;
	ulong          keyChar;
    ulong          newKeyChar = B_ERROR;
	CalcWindow    *calcWindow = NULL;
	
	
	if ( NULL != target )
	{
		calcWindow = cast_as( *target, CalcWindow );
	}
	
	if ( ( NULL != message ) && ( NULL != calcWindow ) )
	{
		keyChar = message->FindLong( "char" );
		if ( B_NO_ERROR <= message->Error() )
		{
		    switch ( keyChar )
		    {
		        case ( CALC_BUTTON_ZERO ):
		        case ( CALC_BUTTON_ONE ):
		        case ( CALC_BUTTON_TWO ):
		        case ( CALC_BUTTON_THREE ):
		        case ( CALC_BUTTON_FOUR ):
		        case ( CALC_BUTTON_FIVE ):
		        case ( CALC_BUTTON_SIX ):
		        case ( CALC_BUTTON_SEVEN ):
		        case ( CALC_BUTTON_EIGHT ):
		        case ( CALC_BUTTON_NINE ):
		        case ( CALC_BUTTON_NEGATE ):
		        case ( CALC_BUTTON_ADD ):
		        case ( CALC_BUTTON_SUBTRACT ):
		        case ( CALC_BUTTON_MULTIPLY ):
		        case ( CALC_BUTTON_MEMORYADD      ):
		        case ( CALC_BUTTON_MEMORYSUBTRACT ):
		            newKeyChar = keyChar;
		        break;
		
		        case ( CALC_BUTTON_EQUALS ):
		        case ( CALC_KEY_ENTER ):
		            newKeyChar = CALC_BUTTON_EQUALS;
		        break;
		
		        case ( CALC_BUTTON_DECIMAL ):
		        case ( CALC_KEY_COMMA ):
		            newKeyChar = CALC_BUTTON_DECIMAL;
		        break;
		
		        case ( CALC_BUTTON_DIVIDE ):
		        case ( CALC_KEY_DIVIDE ):
		            newKeyChar = CALC_BUTTON_DIVIDE;
		        break;
		
		        case ( CALC_BUTTON_CLEARALL ):
		        case ( CALC_KEY_CLEARALL ):
		            newKeyChar = CALC_BUTTON_CLEARALL;
		        break;
		
		        case ( CALC_BUTTON_CLEARENTRY ):
		        case ( CALC_KEY_CLEARENTRY ):
		            newKeyChar = CALC_BUTTON_CLEARENTRY;
		        break;
		
		        case ( CALC_BUTTON_BACKSPACE ):
		        case ( CALC_KEY_BACKSPACE ):
		            newKeyChar = CALC_BUTTON_BACKSPACE;
		        break;
		
		        case ( CALC_BUTTON_MEMORYCLEAR ):
		        case ( CALC_KEY_MEMORYCLEAR ):
		            newKeyChar = CALC_BUTTON_MEMORYCLEAR;
		        break;
		
		        case ( CALC_BUTTON_MEMORYRECALL ):
		        case ( CALC_KEY_MEMORYRECALL ):
		            newKeyChar = CALC_BUTTON_MEMORYRECALL;
		        break;
		
		        default:
		        break;
		    }

			// Ask the window to forward the key to the app.
		    if ( ( B_ERROR != newKeyChar ) )
		    {
				if ( B_NO_ERROR <= calcWindow->forward( newKeyChar ) )
				{
					// If we're successful, the keypress message can be skipped. 
					result = B_SKIP_MESSAGE;
				}
		    }
		}
	}
	
	return( result );
}


Display :: Display( BRect frame, char *name ) : BStringView( frame, name, "0", B_FOLLOW_NONE, B_WILL_DRAW )
{
    SetFontName( CALC_NAMEOF_FONT );
    SetFontSize( CALC_WIDTHOF_FONT );
    SetAlignment( B_ALIGN_RIGHT );
}


Background :: Background( BRect frame, char *name ) : BView( frame, name, B_FOLLOW_NONE, B_WILL_DRAW )
{
	SetViewColor( (uchar)0, (uchar)0, (uchar)0 );
}
