/*
    Calculator.cpp

    Copyright  1996 Levin-Delson Software.  All Rights Reserved.

    $Header: /Bedazzle/dvlp/Calculator/RCS/Calculator.cpp,v 1.15 1996/09/17 18:42:53 adamld Exp $
 */

#include "Calculator.h"

/*
    The buttonLabels and buttonMessages arrays define the layout of the user interface.
    The constants for the number of buttons per row and column are used to arrange the buttons.
*/
char *buttonLabels[ CALC_NUMBEROF_BUTTONS ] =
{
    CALC_NAMEOF_SEVEN,  CALC_NAMEOF_EIGHT, CALC_NAMEOF_NINE,    CALC_NAMEOF_CLEARALL, CALC_NAMEOF_CLEARENTRY, CALC_NAMEOF_MEMORYCLEAR,
    CALC_NAMEOF_FOUR,   CALC_NAMEOF_FIVE,  CALC_NAMEOF_SIX,     CALC_NAMEOF_SUBTRACT, CALC_NAMEOF_DIVIDE,     CALC_NAMEOF_MEMORYSUBTRACT,
    CALC_NAMEOF_ONE,    CALC_NAMEOF_TWO,   CALC_NAMEOF_THREE,   CALC_NAMEOF_ADD,      CALC_NAMEOF_MULTIPLY,   CALC_NAMEOF_MEMORYADD,
    CALC_NAMEOF_NEGATE, CALC_NAMEOF_ZERO,  CALC_NAMEOF_DECIMAL, CALC_NAMEOF_EQUALS,   CALC_NAMEOF_BACKSPACE,  CALC_NAMEOF_MEMORYRECALL
};

ulong buttonMessages[ CALC_NUMBEROF_BUTTONS ] =
{
    CALC_BUTTON_SEVEN,  CALC_BUTTON_EIGHT, CALC_BUTTON_NINE,    CALC_BUTTON_CLEARALL, CALC_BUTTON_CLEARENTRY, CALC_BUTTON_MEMORYCLEAR,
    CALC_BUTTON_FOUR,   CALC_BUTTON_FIVE,  CALC_BUTTON_SIX,     CALC_BUTTON_SUBTRACT, CALC_BUTTON_DIVIDE,     CALC_BUTTON_MEMORYSUBTRACT,
    CALC_BUTTON_ONE,    CALC_BUTTON_TWO,   CALC_BUTTON_THREE,   CALC_BUTTON_ADD,      CALC_BUTTON_MULTIPLY,   CALC_BUTTON_MEMORYADD,
    CALC_BUTTON_NEGATE, CALC_BUTTON_ZERO,  CALC_BUTTON_DECIMAL, CALC_BUTTON_EQUALS,   CALC_BUTTON_BACKSPACE,  CALC_BUTTON_MEMORYRECALL
};

/*
    The decimal point character is kept in a variable.
    It could be changed (e.g. to the European ',') but
    this would require either replacement scanf/sprintf
    routines or a pre- and post-processing search and replace.
*/
char decimalChar = '.';


#pragma mark ## MAIN

int main( void )
{
	int result = B_ERROR;
    Calculator *calculator;


    calculator = new Calculator();
    if ( NULL != calculator )
    {
        (void)calculator->Run();

        delete( calculator );

		result = B_NO_ERROR;
    }

	return( (int)result );
}


#pragma mark ## APPLICATION

/*
    The Calculator class creator has several tasks.

    If other instantiations of Calculator are running, we want to differentiate our window
    (e.g. "Calculator 2").  Query the other instantiations to find out their number.

    We'll turn the Num Lock on each time our window is activated and restore each time deactivated.
    That's done in the Window class, but we need to know the original state of Num Lock so it can
    be reset after _all_ instantiations of Calculator exit.  The first instantiation of Calculator
    saves the state, then all later instantiations "learn" it so they can reset it if they're the
    "last one out".

    The window and all its views are created.  The CalcWindow consists of:
        A background BView, solely for retaining a background color.
        A display BStringView, which presents the numerical display to the user.
            The display's text is updated by the window in response to messages from the Calculator app.
        Many BButtons, each of which sends a unique message to the Calculator app when pressed.

    The window is positioned under the user's cursor (though it's forced to appear completely on-screen).
*/
Calculator :: Calculator() : BApplication( CALC_SIGNATURE )
{
           bool            success = FALSE;
           BRect           calcFrame = BRect( 0, 0, CALC_WIDTHOF_GAP + ( CALC_BUTTONS_PER_ROW * ( CALC_WIDTHOF_BUTTON + CALC_WIDTHOF_GAP ) ), CALC_HEIGHTOF_GAP + ( CALC_BUTTONS_PER_COLUMN * ( CALC_HEIGHTOF_BUTTON + CALC_HEIGHTOF_GAP ) ) + CALC_HEIGHTOF_GAP + CALC_HEIGHTOF_FONT );
           BRect           displayFrame = BRect( CALC_WIDTHOF_GAP, CALC_HEIGHTOF_GAP, calcFrame.right - CALC_WIDTHOF_GAP, CALC_HEIGHTOF_GAP + CALC_HEIGHTOF_FONT + 1 );
           int             button;
           ulong           left;
           ulong           top;
           ulong           widthButton = CALC_WIDTHOF_BUTTON;
           ulong           heightButton = CALC_HEIGHTOF_BUTTON;
           ulong           widthGap = CALC_WIDTHOF_GAP;
           ulong           heightGap = CALC_HEIGHTOF_GAP;
           Display        *display;
           BButton        *bButton;
           Background     *background;
           BList          *calculatorList;
           char            calcTitle[ sizeof( CALC_NAMEOF_WINDOW ) + 5 ];   // Room for " 999\0".
           BPoint          mousePoint = BPoint( 100, 100 );
           ulong           mouseButtons;
           screen_info     screenInfo;
           BRect          *screenRect;
           ulong           marginH, marginV;
    struct CalcData        calcData;


    // Initialize the object's attributes.
    mode = CALC_MODE_OPERAND1 | CALC_MODE_FIRST;
    (void) strcpy( fullString, " 0" );      // fullString's first character holds the minus sign.
    string = &fullString[ 1 ];
    digits = 0;
    operand1 = 0.0;
    memory = 0.0;
    thisCalculator = 1UL;


    /* Query any other instantiations of this app to determine:
        1. A unique number for our window title.
           (It will be one higher than the current highest number.)
        2. The original state of the Num Lock key when the first
           Calculator was started; the last one out will set it back.
    */
    calculatorList = new BList();
    if ( NULL != calculatorList )
    {
        be_roster->GetAppList( CALC_SIGNATURE, calculatorList );
        if ( 1L < calculatorList->CountItems() )    // Don't bother if we're the only Calculator.
        {
            /*
                BList's DoForEach method only allows us to pass one extra parameter to the function performed.
                The CalcData structure is used to contain the two items we need to communicate.
            */
            calcData.cd_ThisCalculator = thisCalculator;
            calcData.cd_OriginalNumLocked = 0UL;
            calculatorList->DoForEach( queryAppCalcData, &calcData );
            thisCalculator = calcData.cd_ThisCalculator;

            // Limit the size of the number to prevent over-running the window title.
            if ( CALC_NUMBEROF_INSTANTIATIONS < thisCalculator )
            {
                thisCalculator = CALC_NUMBEROF_INSTANTIATIONS;
            }

        }
        delete( calculatorList );
    }

    // If we're the only one, we will create the NumLocked value.
    if ( 1UL == thisCalculator )
    {
        originalNumLocked = modifiers() & B_NUM_LOCK;
        (void) strcpy( calcTitle, CALC_NAMEOF_WINDOW );
    }
    else
    {
        originalNumLocked = calcData.cd_OriginalNumLocked;
        (void) sprintf( calcTitle, "%s %ld", CALC_NAMEOF_WINDOW, thisCalculator );
    }

    calcWindow = new CalcWindow( calcFrame, calcTitle, cast_as( be_app, BHandler ) );
    if ( NULL != calcWindow )
    {
        // Cache a connection to the window for future message sending.
        windowMessenger = new BMessenger( calcWindow );
        if ( NULL != windowMessenger )
        {
            /*
                The background view is used to give the window color.  To ensure that it appears behind
                all other views, it will become a child of the window and all other views will become
                children of the background.
            */
            calcFrame.OffsetTo( B_ORIGIN );
            background = new Background( calcFrame, CALC_NAMEOF_BACKGROUND );
            if ( NULL != background )
            {
                calcWindow->AddChild( background );

                // The display view will hold the calculator's numeric display.
                display = new Display( displayFrame, CALC_NAMEOF_DISPLAY );
                if ( NULL != display )
                {
                    background->AddChild( display );

                    /*
                        The buttons are positioned sequentially using the declared arrays and positioning
                        offsets.  Note that the app is specified as the target handler for the button's messages.
                    */
                    for ( button = 0, success = TRUE, left = widthGap, top = displayFrame.bottom + heightGap; ( CALC_NUMBEROF_BUTTONS > button ) && ( TRUE == success ); button++ )
                    {
                        bButton = new BButton( BRect( left, top, left + widthButton, top + heightButton ), buttonLabels[ button ], buttonLabels[ button ], new BMessage( buttonMessages[ button ] ), B_FOLLOW_NONE, B_WILL_DRAW );
                        if ( NULL == bButton )
                        {
                            success = FALSE;
                        }
                        else
                        {
                            bButton->SetTarget( cast_as( be_app, BHandler ) );
                            background->AddChild( bButton );
                        }

                        left += widthButton + widthGap;
                        if ( 0 == ( ( button + 1 ) % CALC_BUTTONS_PER_ROW ) )
                        {
                            left = widthGap;
                            top += heightButton + heightGap;
                        }
                    }

                    if ( TRUE == success )
                    {
                        // Once the program has been debugged, discipline can be relaxed.
                        calcWindow->SetDiscipline( FALSE );

                        /*
                            Position the window under the user's mouse cursor, adjusting to
                            allow the entire window to appear on-screen.
                        */
                        if ( TRUE == calcWindow->Lock() )
                        {
                            background->GetMouse( &mousePoint, &mouseButtons, FALSE );
                            background->ConvertToScreen( &mousePoint );
                            calcWindow->Unlock();
                        }
                        get_screen_info( &screenInfo );
                        screenRect = &screenInfo.frame;

                        marginH = ( screenRect->right - screenRect->left + 1 ) * CALC_MARGINPERCENT_H;
                        marginV = ( screenRect->bottom - screenRect->top  + 1 ) * CALC_MARGINPERCENT_V;

                        mousePoint.ConstrainTo( BRect( screenRect->left + marginH,
                                                       screenRect->top + marginV,
                                                       screenRect->right - marginH - ( calcFrame.right - calcFrame.left + 1),
                                                       screenRect->bottom - marginV - ( calcFrame.bottom - calcFrame.top + 1)
                                                     )
                                               );
                        calcWindow->MoveTo( mousePoint );

                        // Disable the Memory Recall button.
                        enableButton( windowMessenger, CALC_NAMEOF_MEMORYRECALL, FALSE );

                        calcWindow->Show();
                    }
                }
            }
        }
    }
}


/*
    The destructor resets Num Lock to its original state if this is the only instantiation
    of Calculator.  It then disposes of all dynamically-allocated resources which haven't
	been deleted already.  The only way to have quit Calculator was to have either clicked
	in the close button or selected "Quit" from the menu.  In both cases, the Window Looper's
	Quit function was called immediately after the application was informed of the Quit event.
	The Window's Looper disposed of all objects attached to the	window, so we don't have to.
*/
Calculator :: ~Calculator()
{
    BList      *calculatorList;
	
	// Dispose of our BMessenger to the window.
    if ( NULL != windowMessenger )
    {
        delete( windowMessenger );
		windowMessenger = NULL;
    }

    // If we're the only instantiation of Calculator, reset the original state of the Num Lock key.
    calculatorList = new BList();
    if ( NULL != calculatorList )
    {
        be_roster->GetAppList( CALC_SIGNATURE, calculatorList );
        if ( 1L == calculatorList->CountItems() )
        {
            set_keyboard_locks( originalNumLocked | ( modifiers() & ( B_CAPS_LOCK | B_SCROLL_LOCK ) ) );
        }

        delete( calculatorList );
		calculatorList = NULL;
    }
}


void Calculator :: AboutRequested( void )
{
    BAlert *bAlert;
	
	// Note that the revision number is updated automatically by RCS (the dollar symbols are delimiters used by RCS to locate keywords).
    bAlert = new BAlert( "CalculatorInfo", "Calculator " CALC_REVISION " \nCopyright  1996 Levin-Delson Software", "OK");
    if ( NULL != bAlert )
    {
        bAlert->Go();
    }
}


/*
    MessageReceived is the engine of Calculator.  It receives several types of messages:

        Button presses.

        Simulated button presses from the window in response to keyboard input.

        Simulated button presses from the Paste algorithm.

        Requests from new instantiations of Calculator to find out our instantiation number and
        the original state of Num Lock.

        System Clipboard requests (which were received by the window and forwarded).


    The engine is a state machine that accepts input in the form of messages and performs
    the requested actions (inputting numbers, math, memory functions, etc.).
	A simplified illustration of the state machine can be found in the file "CalculatorStateDiagram".
    The class' mode variable holds the state of the machine.

    As numbers are input, they are maintained in a character array.  When it is time to perform
    an operation, the string is converted to a long float (double) and the result is converted
    back into a string for display.  This greatly simplified the input process and display.
*/
void Calculator :: MessageReceived( BMessage *bMessage )
{
    ulong     button;
    double    result;
    double    operand2;
    bool      update = FALSE;
    BMessage *windowMessage;
    BMessage *replyMessage;
    char     *pasteString;
    long      pasteLength;
	bool      isPasted;
	

    if ( NULL != bMessage )
    {
		// This button may have been "pasted" to us. Remember if it was and then remove the PASTE flag.
		isPasted = CALC_MODE_PASTE == ( bMessage->what & CALC_MODE_PASTE );
        button = bMessage->what & ~CALC_MODE_PASTE;

        switch ( button )
        {
            case ( CALC_MESSAGE_QUERY ):
                /* Reply to the message with two answers:
                    1. Which instantiation of Calculator I am.
                    2. The original state of the Num Lock key.
                */
                replyMessage = new BMessage( CALC_MESSAGE_REPLY );
                if ( NULL != replyMessage )
                {
                    if ( B_NO_ERROR <= replyMessage->AddLong( CALC_MESSAGEID_THISCALC, thisCalculator ) )
                    {
						if ( B_NO_ERROR <= replyMessage->AddLong( CALC_MESSAGEID_NUMLOCK, originalNumLocked ) )
                    	{
							if ( B_NO_ERROR <= bMessage->SendReply( replyMessage ) )
							{
								replyMessage = NULL;
							}
						}
					}
					
					// Only delete the message if it _wasn't_ sent. */
					if ( NULL != replyMessage )
					{
						delete( replyMessage );
					}
                }
            break;

            case ( B_PASTE ):
                /*
                    Send each character in the paste string (in its own message) to myself.
                    Each message will be processed as if it were from the appropriate button.
					The PASTE flag is used to allow the '-' character to mean "negative"
					when pasting.  (Search for CALC_BUTTON_SUBTRACT below.)
                */
                if ( CALC_MODE_ERROR != mode )
                {
                    be_clipboard->Lock();

                    pasteString = (char *)be_clipboard->FindText( &pasteLength );
                    while ( 0 < pasteLength-- )
                    {
                        (void)PostMessage( CALC_MODE_PASTE | ( (ulong) *pasteString++ ) );
                    }

                    be_clipboard->Unlock();
                }
            break;

            case ( B_COPY ):
                // Clipboard copy consists of copying the current display string.
                if ( CALC_MODE_ERROR != mode )
                {
                    be_clipboard->Lock();
                    be_clipboard->Clear();

                    be_clipboard->AddText( fullString );

                    be_clipboard->Commit();
                    be_clipboard->Unlock();
                }
            break;

            case ( B_CUT ):
                // Clipboard cut consists of copying the current display string, then clearing entry..
                if ( CALC_MODE_ERROR != mode )
                {
                    be_clipboard->Lock();
                    be_clipboard->Clear();

                    be_clipboard->AddText( fullString );

                    be_clipboard->Commit();
                    be_clipboard->Unlock();
                }
            // Note intentional fall-through of B_CUT.
            case ( CALC_BUTTON_CLEARENTRY ):
                // Reset the display and change the mode to include "first".
                if ( CALC_MODE_ERROR != mode )
                {
                    digits = 0;
                    (void) strcpy( fullString, " 0" );
					mode |= CALC_MODE_FIRST;
                    update = TRUE;
                }
            break;

            case ( CALC_BUTTON_CLEARALL ):
                // Reset the display and set the mode to the initial input state.
                digits = 0;
                (void) strcpy( fullString, " 0" );
                operand1 = 0.0;
                mode = CALC_MODE_OPERAND1 | CALC_MODE_FIRST;
                update = TRUE;
            break;

            case ( CALC_BUTTON_NEGATE ):
                // Negate the display and (if appropriate) the internal variable.
                switch ( mode )
                {
                    case ( CALC_MODE_OPERAND1 | CALC_MODE_FIRST ):
                    case ( CALC_MODE_OPERAND1 ):
                        operand1 = -operand1;

                    // Note intentional fall-through of CALC_MODE_OPERAND1.
                    case ( CALC_MODE_OPERAND2 ):
                        if ( 0 < digits )
                        {
                            if ( ' ' == fullString[ 0 ] )
                            {
                                fullString[ 0 ] = '-';
                            }
                            else
                            {
                                fullString[ 0 ] = ' ';
                            }

                            update = TRUE;
                        }
                    break;

                    case ( CALC_MODE_OPERAND2 | CALC_MODE_FIRST ):
                    case ( CALC_MODE_ERROR ):
                    default:
                    break;
                }
            break;

            case ( CALC_BUTTON_BACKSPACE ):
                // As long as there are characters to remove, remove them.
                switch ( mode )
                {
                    case ( CALC_MODE_OPERAND1 ):
                    case ( CALC_MODE_OPERAND2 ):
                        if ( 0 < digits )
                        {
                            string[ --digits ] = '\0';
                            // Backspacing over all digits removes any negative sign.
                            if ( 0 == digits )
                            {
                                strcpy( fullString, " 0" );
								mode |= CALC_MODE_FIRST;
                            }

                            update = TRUE;
                        }
                    break;

                    default:
                    break;
                }
            break;

            case ( CALC_BUTTON_MEMORYCLEAR ):
                memory = 0.0;
                enableButton( windowMessenger, CALC_NAMEOF_MEMORYRECALL, FALSE );
            break;

            case ( CALC_BUTTON_MEMORYRECALL ):
                /*
                    Recalling memory consists of converting the internal memory value to a string and
                    copying it to the display string.  Note that (when you use the calculator) after
                    recall it is possible to append or remove characters.  It would require an extra
                    state modifier for the state machine to prevent this minor oddity; the extra
                    complexity was not considered worthwhile.
                */
                if ( CALC_MODE_ERROR != mode )
                {
                    digits = dtostr( memory, fullString );
                    if ( CALC_DISPLAY_DIGITS < digits )
                    {
                        (void)strcpy( fullString, "Overflow" );
                        mode = CALC_MODE_ERROR;
                    }
                    else
                    {
                        // Remove the "FIRST" state modifier.
                        mode &= ~CALC_MODE_FIRST;

                        // Prevent the future possibility of a leading zero.
                        if ( ( 1 == digits ) && ( '0' == string[ 0 ] ) )
                        {
                            digits = 0;
                        }
                    }

                    update = TRUE;
                }
            break;

            case ( CALC_BUTTON_DECIMAL ):
            case ( CALC_BUTTON_ZERO ):
            case ( CALC_BUTTON_ONE ):
            case ( CALC_BUTTON_TWO ):
            case ( CALC_BUTTON_THREE ):
            case ( CALC_BUTTON_FOUR ):
            case ( CALC_BUTTON_FIVE ):
            case ( CALC_BUTTON_SIX ):
            case ( CALC_BUTTON_SEVEN ):
            case ( CALC_BUTTON_EIGHT ):
            case ( CALC_BUTTON_NINE ):
                if ( CALC_MODE_ERROR != mode )
                {
                    if ( CALC_MODE_FIRST == ( mode & CALC_MODE_FIRST ) )
                    {
						if ( 0 == digits )
						{
                         	(void) strcpy( string, "0" );
						}
						else
						{						
                        	digits = 0;
                        	(void) strcpy( fullString, " 0" );
						}
                        mode &= ~CALC_MODE_FIRST;
                    }

                    if ( CALC_DISPLAY_DIGITS > digits )
                    {
                        // If this would be a leading zero, skip the request.
                        if ( ( CALC_BUTTON_ZERO != button ) || ( 0 < digits ) )
                        {
	                        // If the button is a decimal point but there's already one present, skip the request.
	                        if ( ( CALC_BUTTON_DECIMAL != button ) || ( NULL == strchr( string, decimalChar ) ) )
	                        {
		                        string[ digits++ ] = (char)button;
		                        string[ digits ] = '\0';
							}
						}
                    }
	                update = TRUE;
                }
            break;

            case ( CALC_BUTTON_MEMORYADD ):
            case ( CALC_BUTTON_MEMORYSUBTRACT ):
            case ( CALC_BUTTON_EQUALS ):
            case ( CALC_BUTTON_ADD ):
            case ( CALC_BUTTON_SUBTRACT ):
            case ( CALC_BUTTON_MULTIPLY ):
            case ( CALC_BUTTON_DIVIDE ):
                switch ( mode )
                {
                    case ( CALC_MODE_OPERAND1 ):
                        if ( 0 < digits )
                        {
                            // Convert the display string to a double and save it as operand1.
                            operand1 = strtod( fullString, (char **)NULL );
                        }
                        else
                        {
                            operand1 = 0.0;
                        }
                    // Note intentional fall-through of CALC_MODE_OPERAND1.
                    case ( CALC_MODE_OPERAND1 | CALC_MODE_FIRST ):
                    case ( CALC_MODE_OPERAND2 | CALC_MODE_FIRST ):
                        if ( CALC_MODE_ERROR != mode )
                        {
                            switch ( button )
                            {
                                case ( CALC_BUTTON_MEMORYADD ):
                                    memory += operand1;
                                    enableButton( windowMessenger, CALC_NAMEOF_MEMORYRECALL, TRUE );
                                    mode |= CALC_MODE_FIRST;
                                break;

                                case ( CALC_BUTTON_MEMORYSUBTRACT ):
                                    memory -= operand1;
                                    enableButton( windowMessenger, CALC_NAMEOF_MEMORYRECALL, TRUE );
                                    mode |= CALC_MODE_FIRST;
                                break;

                                case ( CALC_BUTTON_EQUALS ):
									if ( ( CALC_MODE_FIRST | CALC_MODE_OPERAND2 ) == mode )
									{
                                    	mode = CALC_MODE_FIRST | CALC_MODE_OPERAND1;
									}
									else
									{
                                    	operation = CALC_BUTTON_EQUALS;
                                    	mode = CALC_MODE_OPERAND2 | CALC_MODE_FIRST;
									}
									
                                break;

								case ( CALC_BUTTON_SUBTRACT ):
									/*
										If this button is being "pasted" in, and we're still in a "first"
										mode, allow '-' to mean "negative".  This gives the ability to 
										paste negative numbers.  Try pasting this, "3.14 * -2 - -3.14 =".
									*/
									if ( ( TRUE == isPasted ) && CALC_MODE_FIRST == ( mode & CALC_MODE_FIRST ) )
									{
                    					if ( CALC_MODE_OPERAND1 == ( mode & CALC_MODE_OPERAND1 ) )	
										{
                       						operand1 = -operand1;
										}
										
										if ( 0 <= digits )
										{
											digits = 0;
											(void)strcpy( fullString, " 0" );
										}
										
			                            if ( ' ' == fullString[ 0 ] )
			                            {
			                                fullString[ 0 ] = '-';
			                            }
			                            else
			                            {
			                                fullString[ 0 ] = ' ';
			                            }
										
										update = TRUE;
									} 
									else
									{
										operation = CALC_BUTTON_SUBTRACT;
	                                    mode = CALC_MODE_OPERAND2 | CALC_MODE_FIRST;
									}
								break;
								
                                default:
                                    operation = button;
                                    mode = CALC_MODE_OPERAND2 | CALC_MODE_FIRST;
                                break;
                            }
                        }
                    break;

                    case ( CALC_MODE_OPERAND2 ):
                        if ( 0 < digits )
                        {
                            // Convert the display string to a double and save as operand2.
                            operand2 = strtod( fullString, (char **)NULL );
                        }
                        else
                        {
                            operand2 = 0.0;
                        }

                        switch ( operation )
                        {
                            case ( CALC_BUTTON_ADD ):
                                result = operand1 + operand2;
                            break;

                            case ( CALC_BUTTON_SUBTRACT ):
                                result = operand1 - operand2;
                            break;

                            case ( CALC_BUTTON_MULTIPLY ):
                                result = operand1 * operand2;
                            break;

                            case ( CALC_BUTTON_DIVIDE ):
                                // Avoid a divide-by-zero error.
                                if ( 0.0 == operand2 )
                                {
                                    (void)strcpy( fullString, "Divide by 0" );
                                    mode = CALC_MODE_ERROR;
                                    update = TRUE;
                                }
                                else
                                {
                                    result = operand1 / operand2;
                                }
                            break;

                            case ( CALC_MODE_ERROR ):
                            default:
                            break;
                        }

                        if ( CALC_MODE_ERROR != mode )
                        {
                            // Convert the result to a string and copy it to the display string.
                            digits = dtostr( result, fullString );
                            if ( CALC_DISPLAY_DIGITS < digits )
                            {
                                (void)strcpy( fullString, "Overflow" );
                                mode = CALC_MODE_ERROR;
                            }
                            else
                            {
                                operand1 = result;
                                switch ( button )
                                {
                                    case ( CALC_BUTTON_MEMORYADD ):
                                        memory += result;
                                        enableButton( windowMessenger, CALC_NAMEOF_MEMORYRECALL, TRUE );
                                        mode = CALC_MODE_OPERAND1 | CALC_MODE_FIRST;
                                    break;

                                    case ( CALC_BUTTON_MEMORYSUBTRACT ):
                                        memory -= result;
                                        enableButton( windowMessenger, CALC_NAMEOF_MEMORYRECALL, TRUE );
                                        mode = CALC_MODE_OPERAND1 | CALC_MODE_FIRST;
                                    break;

                                    case ( CALC_BUTTON_EQUALS ):
                                        mode = CALC_MODE_OPERAND1 | CALC_MODE_FIRST;
                                    break;

                                    default:
                                        mode = CALC_MODE_OPERAND2 | CALC_MODE_FIRST;
                                        operation = button;
                                    break;
                                }
                            }

                            update = TRUE;
                        }
                    break;

                    case ( CALC_MODE_ERROR ):
                    default:
                    break;
                }
            break;

            default:
                // If the message isn't one we handle, pass it to the parent class.
                BApplication::MessageReceived( bMessage );
            break;
        }

        /*
            Updating the window's display requires sending a message containing the display string
            to the window's handler.  Only do this if an update is necessary.
        */
        if ( TRUE == update )
        {
            windowMessage = new BMessage( CALC_MESSAGE_SETTEXT );
            if ( NULL != windowMessage )
            {
                if ( B_NO_ERROR <= windowMessage->AddString( CALC_MESSAGEID_OBJECTNAME, CALC_NAMEOF_DISPLAY ) )
                {
	                if ( B_NO_ERROR <= windowMessage->AddString( CALC_MESSAGEID_TEXT, fullString ) )
	                {
	                    if ( B_NO_ERROR <= windowMessenger->SendMessage( windowMessage ) )
						{
							windowMessage = NULL;
						}
	                }
				}
				
				// Only delete the message if it _wasn't_ sent. */
				if ( NULL != windowMessage )
				{
					delete( windowMessage );
				}
            }
        }
    }
}


/*
    QueryAppCalcData asks the "TeamID" instantiation of Calculator for its instantiation number
    and the original state of the Num Lock key.  It is called by BList's DoForEach() method.
    Returns TRUE (meaning "stop walking the BList") only if invalid parameters are passed in.
*/
bool queryAppCalcData( void *voidTeamID, void *voidCalcData )
{
           bool        result = TRUE;
           team_id     teamID;
    struct CalcData   *calcData;
           app_info    myAppInfo;
           BMessenger *queryMessenger;
           BMessage   *replyMessage = NULL;
           ulong       replyNumber;


    if ( ( NULL != voidTeamID ) && ( NULL != voidCalcData ) )
    {
        result = FALSE;
        teamID = (team_id)voidTeamID;
        calcData = (struct CalcData *)voidCalcData;

        // Find out my team ID and don't bother querying myself.
        if ( B_NO_ERROR <= be_app->GetAppInfo( &myAppInfo ) )
        {
            if ( myAppInfo.team != teamID )
            {
                // Create a messenger to the app.  This could fail if the app has quit.
                queryMessenger = new BMessenger( myAppInfo.signature, teamID );
                if ( NULL != queryMessenger )
                {
                    // Send a synchronous message to the app.
                    if ( B_NO_ERROR <= queryMessenger->SendMessage( CALC_MESSAGE_QUERY, &replyMessage ) )
                    {
                        if ( NULL != replyMessage )
                        {
                            if ( B_NO_REPLY != ( replyMessage->what ) )
                            {
                                replyNumber = replyMessage->FindLong( CALC_MESSAGEID_THISCALC );
                                if ( B_NO_ERROR <= replyMessage->Error() )
                                {
                                    // Insure that the passed number is greater than the queried App's number.
                                    if ( replyNumber >= calcData->cd_ThisCalculator )
                                    {
                                        calcData->cd_ThisCalculator = replyNumber + 1UL;
                                    }
                                }

                                replyNumber = replyMessage->FindLong( CALC_MESSAGEID_NUMLOCK );
                                if ( B_NO_ERROR <= replyMessage->Error() )
                                {
                                    calcData->cd_OriginalNumLocked = replyNumber;
                                }
                            }

                            // It is our responsibility to dispose of the reply message.
                            delete( replyMessage );
                        }
                    }
                }
            }
        }
    }

    return ( result );
}


/*
    Converts a double to string, trimming trailing zeros and decimal point.  Returns the number of digits.
*/
int dtostr( double result, char *fullString )
{
    int   digits;
    char *string;


    if ( NULL != fullString )
    {
        string = &fullString[ 1 ];

        // Handle negative values by making them positive and inserting a '-' into the string.
        if ( 0 > result )
        {
            fullString[ 0 ] = '-';
            result = -result;
        }
        else
        {
            fullString[ 0 ] = ' ';
        }

        digits = sprintf( string, "%Lf", result );  // Note type modifier for long float (double).

        // If there are more than CALC_DISPLAY_DIGITS significant digits, don't bother trying to fit it.
        if ( CALC_DISPLAY_DIGITS > (int)( strchr( string, decimalChar ) - string ) )
        {
            // Truncate trailing zeros.
            while ( ( 0 < digits ) && ( '0' == string[ digits - 1 ] ) )
            {
                string[ --digits ] = '\0';
            }

            // Truncate trailing decimal point.
            if ( ( 0 < digits ) && ( decimalChar == string[ digits - 1 ] ) )
            {
                string[ --digits ] = '\0';
            }

            // If there are still too many digits...
            if ( CALC_DISPLAY_DIGITS <= digits )
            {
                // Truncate leading zero.
                if ( '0' == string[ 0 ] )
                {
                    strcpy( string, &string[ 1 ] );
                    digits--;
                }

                // Then truncate the string.
                digits = min( digits, CALC_DISPLAY_DIGITS );
                string[ digits ] = '\0';

                // We may again have trailing zeros to truncate.
                while ( ( 0 < digits ) && ( '0' == string[ digits - 1 ] ) )
                {
                    string[ --digits ] = '\0';
                }
                // And another chance to truncate a trailing decimal point.
                if ( ( 0 < digits ) && ( decimalChar == string[ digits - 1 ] ) )
                {
                    string[ --digits ] = '\0';
                }

                if ( 0 >= digits )
                {
                    digits = 1;
                    strcpy( fullString, " 0" );
                }
            }
        }
    }
    return ( digits );
}


// Send a message to the CalcWindow asking it to enable/disable a button.
void enableButton( BMessenger *bMessenger, char *buttonName, bool enable )
{
    BMessage *bMessage;
    if ( ( NULL != bMessenger ) && ( NULL != buttonName ) )
    {
        bMessage = new BMessage( CALC_MESSAGE_SETENABLED );
        if ( NULL != bMessage )
        {
            if ( B_NO_ERROR <= bMessage->AddString( CALC_MESSAGEID_OBJECTNAME, buttonName ) )
            {
                if ( B_NO_ERROR <= bMessage->AddBool( CALC_MESSAGEID_ENABLED, enable ) )
                {
                    if ( B_NO_ERROR <= bMessenger->SendMessage( bMessage ) )
					{
						bMessage = NULL;
					}
                }
            }
			
			if ( NULL != bMessage )
			{
				// Only delete the message if it _wasn't_ sent. */
				delete( bMessage );
			}
        }
    }
}
